#!/usr/bin/env python

"""
By using changelog file as an input (can be generated with 'generate_changelog.py' script),
this script goes through all packages in changelog file (in order listed), tags them correctly
as per the file, and releases them one by one.
This script uses python-gitlab package for accessing GitLab's API.

This script uses the provided changelog file to release a package.
The changelog is expected to have the following structure:

    * package name
      * tag1 name (date of the tag).
        * tag description. Each line of the tag description starts with `*` character.
        - commits (from earliest to latest). Each line of the commit starts with `-` character.
      * tag2 name (date of the tag).
        * tag description. Each line of the tag description starts with `*` character.
        - commits (from earliest to latest). Each line of the commit starts with `-` character.
      * patch
        - leftover not-tagged commits (from earliest to latest)

This script can also be used to release a single package.
`IMPORTANT`: there are some considerations that needs to be taken
into account **before** you release a new version of a package:

  * In the changelog file:
    - write the name of this package and write the last tag.
      For the tag name, you can either indicate `patch`, `minor` or `major`,
      and the package will be then released with either patch, minor, or major version bump.
    - Alternatively, you can specify the version name directly but be careful that it is higher
      than the last release tag of this package.
      Make sure that the version that you are trying to release is not already released.
      Also, make sure you follow semantic versions: http://semver.org.
    - Then, under the desired new tag version of the package, please write down the changes that are applied
      to the package between the last released version and this version. This
      changes are written to release tags of packages in the Gitlab interface.
      For an example look at: https://gitlab.idiap.ch/bob/bob.extension/tags
  * Make sure all the tests for the package are passing.
  * Make sure the documentation is building with the following command:
    ``sphinx-build -aEWn doc sphinx``
  * Make sure all changes are committed to the git repository and pushed.
  * Make sure the documentation badges in README.rst are pointing to:
    https://www.idiap.ch/software/bob/docs/bob/...
  * For database packages, make sure that the '.sql3' file or other
    metadata files have been generated (if any).
  * Make sure bob.nightlies is green after the changes are submitted if the
    package is a part of the nightlies.
  * If your package depends on an unreleased version of another Bob package,
    you need to release that package first.


Usage:
    {0} [-v...] [options] [--] <private_token>
    {0} -h | --help
    {0} --version

Arguments:
    <private_token>  Private token used to access GitLab.

Options:
    -h --help                     Show this screen.
    --version                     Show version.
    -c, --changelog-file STR      A changelog file with all packages to release with their tags, listed in order.
                                  [default: changelog_since_last_release.md].
    -g, --group-name STR          Group name where we are assuming that all packages are located.
                                  [default: bob].
    -p, --package STR             If the name of a package is provided, then this package will be found
                                  in the changelog file and the release will resume from it (if option --resume is set)
                                  or only this package will be released.
    -r, --resume                  The overall release will resume from the provided package name.
    -q, --dry-run                 Only print the actions, but do not execute them.

"""

import sys
import os
from docopt import docopt
import gitlab
import datetime
import re
from distutils.version import StrictVersion
import numpy
import time


def _insure_correct_package(candidates, group_name, pkg_name):
    for pkg in candidates:
        # make sure the name and the group name match exactly
        if pkg.name == pkg_name and pkg.namespace['name'] == group_name:
            return pkg
    raise ValueError('Package "{0}" was not found inside group "{1}"'.format(pkg_name, group_name))


# adapted from the same-name function in new_version.py script of bob.extension package
def _update_readme(readme, version=None):
    """
    Inside text of the readme, replaces parts of the links to the provided version.
    If version is not provided, replace to `stable` or `master`.
    Args:
        readme: Text of the README.rst file from a bob package
        version: Format of the version string is '#.#.#'

    Returns: New text of readme with all replaces done

    """
    # replace the badge in the readme's text with the given version
    DOC_IMAGE = re.compile(r'\-(stable|(v\d+\.\d+\.\d+([abc]\d+)?))\-')
    BRANCH_RE = re.compile(r'/(stable|master|(v\d+\.\d+\.\d+([abc]\d+)?))')

    new_readme = []
    for line in readme.splitlines():
        if BRANCH_RE.search(line) is not None:
            if "gitlab" in line:  # gitlab links
                replacement = "/v%s" % version if version is not None else "/master"
                line = BRANCH_RE.sub(replacement, line)
            if "software/bob" in line:  # our doc server
                if 'master' not in line:  # don't replace 'latest' pointer
                    replacement = "/v%s" % version if version is not None \
                        else "/stable"
                    line = BRANCH_RE.sub(replacement, line)
        if DOC_IMAGE.search(line) is not None:
            replacement = '-v%s-' % version if version is not None else '-stable-'
            line = DOC_IMAGE.sub(replacement, line)
        new_readme.append(line)
    return '\n'.join(new_readme)


def get_latest_tag_name(gitpkg):
    """
    Find the name of the latest tag for a given package in the format '#.#.#'
    Args:
        gitpkg: gitlab package object

    Returns: The name of the latest tag in format '#.#.#'. None if no tags for the package were found.

    """
    # get 50 latest tags as a list
    latest_tags = gitpkg.tags.list(all=True)
    if not latest_tags:
        return None
    # create list of tags' names but ignore the first 'v' character in each name
    # also filter out non version tags
    tag_names = [tag.name[1:] for tag in latest_tags if StrictVersion.version_re.match(tag.name[1:])]
    # sort them correctly according to each subversion number
    tag_names.sort(key=StrictVersion)
    # take the last one, as it is the latest tag in the sorted tags
    latest_tag_name = tag_names[-1]
    return latest_tag_name


def get_parsed_tag(gitpkg, tag):
    """
        An older tag is formatted as 'v2.1.3 (Sep 22, 2017 10:37)', from which we need only v2.1.3
        The latest tag is either patch, minor, major, or none
    """
    m = re.search(r"(v\d.\d.\d)", tag)
    if m:
        return m.group(0)
    # tag = Version(tag)

    # if we bump the version, we need to find the latest released version for this package
    if 'patch' == tag or 'minor' == tag or 'major' == tag:
        # find the correct latest tag of this package (without 'v' in front),
        # None if there are no tags yet
        latest_tag_name = get_latest_tag_name(gitpkg)
        # if there were no tags yet, assume the very first version
        if not latest_tag_name:
            return 'v0.0.1'
        # check that it has expected format #.#.#
        # latest_tag_name = Version(latest_tag_name)
        m = re.match(r"(\d.\d.\d)", latest_tag_name)
        if not m:
            raise ValueError(
                'The latest tag name {0} in package {1} has unknown format'.format('v' + latest_tag_name, gitpkg.name))
        # increase the version accordingly
        major, minor, patch = latest_tag_name.split('.')
        if 'major' == tag:  # increment the first number in 'v#.#.#' but make minor and patch to be 0
            return 'v' + str(int(major) + 1) + '.0.0'
        if 'minor' == tag:  # increment the second number in 'v#.#.#' but make patch to be 0
            return 'v' + major + '.' + str(int(minor) + 1) + '.0'
        if 'patch' == tag:  # increment the last number in 'v#.#.#'
            return 'v' + major + '.' + minor + '.' + str(int(patch) + 1)

    if 'none' == tag:  # we do nothing in this case
        return tag

    raise ValueError('Cannot parse changelog tag {0} of the package {1}'.format(tag, gitpkg.name))


def update_tag_comments(gitpkg, tag_name, tag_comments_list, dry_run=False):
    """
    Write annotations inside the provided tag of a given package.
    Args:
        gitpkg: gitlab package object
        tag_name: The name of the tag to update
        tag_comments_list: New annotations for this tag in a form of list
        dry_run: If True, nothing will be committed or pushed to GitLab

    Returns: The gitlab object for the tag that was updated

    """
    # get tag and update its description
    tag = gitpkg.tags.get(tag_name)
    print('Found tag {1}, updating its comments with:'.format(gitpkg.name, tag.name))
    print(tag_comments_list)
    if not dry_run:
        tag.set_release_description('\n'.join(tag_comments_list))
    return tag


def commit_files(gitpkg, files_dict, message='Updated files', dry_run=False):
    """
    Commit files of a given GitLab package.
    Args:
        gitpkg: gitlab package object
        files_dict: Dictionary of file names and their contents (as text)
        message: Commit message
        dry_run: If True, nothing will be committed or pushed to GitLab

    """
    data = {
        'branch': 'master',  # v4
        'commit_message': message,
        'actions': []
    }
    # add files to update
    for filename in files_dict.keys():
        update_action = dict(action='update', file_path=filename)
        update_action['content'] = files_dict[filename]
        data['actions'].append(update_action)

    print("Committing changes in files: {0}".format(str(files_dict.keys())))
    if not dry_run:
        gitpkg.commits.create(data)


def get_last_pipeline(gitpkg):
    """
    Returns the last pipeline of the project
    Args:
        gitpkg: gitlab package object

    Returns: The gtilab object of the pipeline

    """
    # wait for 10 seconds to ensure that if a pipeline was just submitted,
    # we can retrieve it
    time.sleep(10)
    # get the last pipeline
    return gitpkg.pipelines.list(per_page=1, page=1)[0]


def just_build_package(gitpkg, dry_run=False):
    """
    Creates the pipeline with the latest tag and starts it
    Args:
        gitpkg: gitlab package object
        dry_run: If True, the pipeline will not be created on GitLab

    Returns:

    """

    # get the latest tag
    latest_tag_name = 'v' + get_latest_tag_name(gitpkg)
    # create the pipeline with this tag and start it
    print("Creating and starting pipeline for tag {0}".format(latest_tag_name))
    if not dry_run:
        new_pipeline = gitpkg.pipelines.create({'ref': latest_tag_name})
        return new_pipeline.id

    return None


def wait_for_pipeline_to_finish(gitpkg, pipeline_id, dry_run=False):
    """
    Using sleep function, wait for the latest pipeline to finish building.
    This function pauses the script until pipeline completes either successfully or with error.
    Args:
        gitpkg: gitlab package object
        pipeline_id: id of the pipeline for which we are waiting to finish
        dry_run: If True, print log message and exit. There wil be no waiting.

    """
    sleep_step = 30
    max_sleep = 120 * 60  # two hours
    # pipeline = get_last_pipeline(gitpkg, before_last=before_last)

    print('Waiting for the pipeline {0} of package {1} to finish. Do not interrupt.'.format(pipeline_id, gitpkg.name))

    if dry_run:
        return

    # retrieve the pipeline we are waiting for
    pipeline = gitpkg.pipelines.get(pipeline_id)

    # probe and wait for the pipeline to finish
    slept_so_far = 0
    while pipeline.status == 'running' or pipeline.status == 'pending':
        time.sleep(sleep_step)
        slept_so_far += sleep_step
        if slept_so_far > max_sleep:
            raise ValueError('I cannot wait longer than {0} seconds for '
                             'pipeline {1} to finish running!'.format(max_sleep, pipeline_id))
        # probe gitlab to update the status of the pipeline
        pipeline = gitpkg.pipelines.get(pipeline_id)

    # finished running, now check if it succeeded
    if pipeline.status != 'success':
        raise ValueError('Pipeline {0} of project {1} exited with undesired status "{2}". '
                         'Release is not possible.'.format(pipeline_id, gitpkg.name, pipeline.status))

    print('Pipeline {0} of package {1} succeeded. Continue processing.'.format(pipeline_id, gitpkg.name))


def cancel_last_pipeline(gitpkg):
    """
    Cancel the last started pipeline of a package
    Args:
        gitpkg: gitlab package object

    """
    pipeline = get_last_pipeline(gitpkg)
    print('Cancelling the last pipeline {0} of project {1}'.format(pipeline.id, gitpkg.name))
    pipeline.cancel()


def release_package(gitpkg, tag_name, tag_comments_list, dry_run=False):
    """
    Release package. The provided tag will be annotated with a given list of comments.
    README.rst and version.txt files will also be updated according to the release procedures.
    Args:
        gitpkg: gitlab package object
        tag_name: The name of the release tag
        tag_comments_list: New annotations for this tag in a form of list
        dry_run: If True, nothing will be committed or pushed to GitLab

    """
    # if there is nothing to release, just rebuild the package
    if tag_name == 'none':
        print("Since the tag is 'none', we just re-build the last pipeline")
        return just_build_package(gitpkg, dry_run)

    # 1. Replace branch tag in Readme to new tag, change version file to new version tag. Add and commit to gitlab
    version_number = tag_name[1:]  # remove 'v' in front
    readme_file = gitpkg.files.get(file_path='README.rst', ref='master')
    readme_content = readme_file.decode().decode()
    readme_content = _update_readme(readme_content, version=version_number)
    # commit and push changes
    commit_files(gitpkg, {'README.rst': readme_content, 'version.txt': version_number},
                 'Increased stable version to %s' % version_number, dry_run)
    if not dry_run:
        # cancel running the pipeline triggered by the last commit
        cancel_last_pipeline(gitpkg)

    # 2. Tag package with new tag and push
    print("Creating tag {}".format(tag_name))
    print("updating tag's comments with:".format(gitpkg.name, tag_name))
    print(tag_comments_list)
    if not dry_run:
        tag = gitpkg.tags.create({'tag_name': tag_name, 'ref': 'master'})
        # update tag with comments
        tag.set_release_description('\n'.join(tag_comments_list))

    # get the pipeline that is actually running with no skips
    running_pipeline = get_last_pipeline(gitpkg)

    # 3. Replace branch tag in Readme to master, change version file to beta version tag. Git add, commit, and push.
    readme_content = _update_readme(readme_content)
    major, minor, patch = version_number.split('.')
    version_number = '{}.{}.{}b0'.format(major, minor, int(patch)+1)
    # commit and push changes
    commit_files(gitpkg, {'README.rst': readme_content, 'version.txt': version_number},
                 'Increased latest version to %s [skip ci]' % version_number, dry_run)
    return running_pipeline.id


def parse_and_process_package_changelog(gl, bob_group, pkg_name, package_changelog, dry_run=False):
    """
    Process the changelog of a single package. Parse the log following specific format.
    Update annotations of the provided older tags and release the package by following the last tag description.
    Args:
        gl: Gitlab API object
        bob_group: gitlab object for the group
        pkg_name: name of the package
        package_changelog: the changelog corresponding to the provided package
        dry_run: If True, nothing will be committed or pushed to GitLab

    Returns: gitlab handle for the package, name of the latest tag, and tag's comments

    """
    cur_tag = None
    cur_tag_comments = []

    grpkg = _insure_correct_package(bob_group.projects.list(search=pkg_name), bob_group.name, pkg_name)
    # so, we need to retrieve the full info from GitLab using correct project id
    gitpkg = gl.projects.get(id=grpkg.id)

    # we assume that changelog is formatted as structured text
    # first line is the name of the package
    for line in package_changelog:
        if '  *' == line[:3]:  # a tag level
            # write the comments collected for the previous tag
            if cur_tag:
                update_tag_comments(gitpkg, cur_tag, cur_tag_comments, dry_run)
                cur_tag_comments = []  # reset comments

            # parse the current tag name
            cur_tag = get_parsed_tag(gitpkg, line[3:].strip())

        else:  # all other lines are assumed to be comments
            cur_tag_comments.append(line.strip())

    # return the last tag and comments for release
    return gitpkg, cur_tag, cur_tag_comments


def main(private_token, group_name='bob', changelog_file='changelog.rst', dry_run=False, package=None, resume=False):
    """
    Main function that updates and releases packages according to the provided changelog file.
    Args:
        private_token: GitLab token with the access rights to update and release Bob's packages
        group_name: Name of the group in GitLab. By default, the name of the group is 'bob'
        changelog_file: The changelog file that defines which packages will be processed
        dry_run: If True, nothing will be committed or pushed to GitLab
        package: If provided, only this package will be processed
        resume: If True, the processing of changelog will be resumed starting with the provided package

    """
    gl = gitlab.Gitlab('https://gitlab.idiap.ch', private_token=private_token, api_version=4)
    bob_group = gl.groups.list(search=group_name)[0]

    # traverse all packages in the changelog, edit older tags with updated comments,
    # tag them with a suggested version, then try to release, and
    # wait until done to proceed to the next package
    with open(changelog_file) as f:
        changelog = f.readlines()

    # find the starts of each package's description in the changelog
    pkgs = numpy.asarray([i for i, line in enumerate(changelog) if line[0] == '*'])
    pkgs = numpy.concatenate([pkgs, [len(changelog)]])
    start_idx = 0
    if package:
        # get the index where the package first appears in the list
        start_idx = [i for i, line in enumerate(changelog) if line[1:].strip() == package]
        if not start_idx:
            print('Package {0} was not found in the changelog'.format(package))
            return
        start_idx = pkgs.tolist().index(start_idx[0])

    for i in range(start_idx, pkgs.shape[0] - 1):
        cur_package_name = changelog[pkgs[i]][1:].strip()
        print('\nProcessing package {0}'.format(changelog[pkgs[i]]))
        gitpkg, tag, tag_comments = parse_and_process_package_changelog(gl, bob_group, cur_package_name,
                                                                        changelog[pkgs[i] + 1: pkgs[i + 1]], dry_run)
        # release the package with the found tag and its comments
        if gitpkg:
            pipeline_id = release_package(gitpkg, tag, tag_comments, dry_run)
            # now, wait for the pipeline to finish, before we can release the next package
            wait_for_pipeline_to_finish(gitpkg, pipeline_id, dry_run)

        # if package name is provided and resume is not set, process only this package
        if package == cur_package_name and not resume:
            break

    print('\nFinished processing changelog')


if __name__ == '__main__':
    arguments = docopt(__doc__.format(sys.argv[0]), version='Changelog 0.0.1')
    main(arguments['<private_token>'], group_name=arguments['--group-name'],
         changelog_file=arguments['--changelog-file'], dry_run=arguments['--dry-run'],
         package=arguments['--package'], resume=arguments['--resume'])
